#include <stdlib.h>      


#define order 200
#define numberoftasks 4
#define grainsize 1000


float crandom(int *seed);
void SubA( float a[order][order],
           float x[order],
           float b[order],
           int *n,
           float r[order],
           float p[order],
           float workold[numberoftasks],
           float worknew[numberoftasks],
           float workpap[numberoftasks],
           float *alpha,
           float *beta,
           int *tasks,
           int *lockid,
           int *eventa,
           int *eventb,
           int *subset1,
           int *subset23,
           int *subset4,
           int *subset56,
           int *subset78,
           int *subset9,
           int *outstanding,
           int *taskindex,
           float *tolerance,
           int *done );
      

      
main()
{
/*
;
;  This is a main function for the example 5. This example, in parallel, 
;  solves a system of linear equations in the form [A]{X}={B}
;  by the conjugate gradient method. This example uses a parallel lock
;  to handle all the critical sections, and uses two parallel events
;  to handle task synchronization.
;  
*/
   static int n, lockid, eventa, eventb, subset1, subset23, subset4;
   static int subset56, subset78, subset9, outstanding, tasks;
   static int taskid[numberoftasks], taskindex[numberoftasks];
   int j4temp, k4temp, seed = 88888888;
   static float a[order][order], x[order], b[order], r[order], p[order];
   static float workold[numberoftasks], worknew[numberoftasks];   
   static float workpap[numberoftasks], alpha, beta, tolerance = 0.1;
   static int done;

/*  generate data in a sequential procedure  */

   printf(">>> Generate data ...\n");
   tasks = numberoftasks;
   n = order;
   for ( j4temp = 0; j4temp < n; j4temp ++ )
      for ( k4temp =0; k4temp < n; k4temp ++ )
            a[k4temp][j4temp] = crandom( &seed );
   for ( j4temp = 0; j4temp < n; j4temp ++ )
   {
      for ( k4temp =0; k4temp < n; k4temp ++ )
            a[j4temp][j4temp] = a[j4temp][j4temp] + a[j4temp][k4temp];
      a[j4temp][j4temp] = a[j4temp][j4temp] * 5.0;
   }
   for (j4temp = 0; j4temp < n; j4temp ++ )
   {
      b[j4temp] = 0.0;
      for ( k4temp =0; k4temp < n; k4temp ++ )
            b[j4temp] = b[j4temp] + a[j4temp][k4temp];
   }

/*  create parallel lock  */

   CreateParallelLock(&lockid);

/*  create parallel events  */

   CreateParallelEvent(&eventa);
   CreateParallelEvent(&eventb);

/*  determine length of subsets  */

   subset1 = __min(grainsize, order/numberoftasks);
   if( subset1 == 0 ) subset1 = 1;
   subset23 = __min(grainsize/4, order/numberoftasks);
   if( subset23 == 0 ) subset23 = 1;
   subset4 = __min(grainsize/3/order, order/numberoftasks);
   if( subset4 == 0 ) subset4 = 1;
   subset56 = __min(grainsize / ( 3 * order + 2), order / numberoftasks);
   if( subset56 == 0 ) subset56 = 1;
   subset78 = __min(grainsize/2, order/numberoftasks);
   if( subset78 == 0 ) subset78 = 1;
   subset9 = __min(grainsize/2, order/numberoftasks);
   if( subset9 == 0 ) subset9 = 1;

/*  initialize  */

   outstanding = 0;
   InitializeParallelEvent(&eventa, &tasks);

/*  solve in parallel  */

   printf(">>> Solve in parallel ...\n");
   for ( j4temp =0; j4temp < numberoftasks - 1; j4temp ++ )
   {
      taskindex[j4temp] = j4temp;
      Dispatch_25( &taskid[j4temp], SubA, a,
                                          x,
                                          b,
                                          &n,
                                          r,
                                          p,
                                          workold,
                                          worknew,
                                          workpap,
                                          &alpha,
                                          &beta,
                                          &tasks,
                                          &lockid,
                                          &eventa,
                                          &eventb,
                                          &subset1,
                                          &subset23,
                                          &subset4,
                                          &subset56,
                                          &subset78,
                                          &subset9,
                                          &outstanding,
                                          &taskindex[j4temp],
                                          &tolerance,
                                          &done);
   }
   taskindex[tasks-1] = tasks-1;
   SubA( a, x, b, &n, r, p, workold, worknew, workpap, &alpha,
         &beta, &tasks, &lockid, &eventa, &eventb, &subset1,
         &subset23, &subset4, &subset56, &subset78, &subset9,
         &outstanding, &taskindex[tasks-1], &tolerance, &done);

/*  terminate tasks  */

   WaitForAndTerminateAllTasks( );

/*  delete parallel lock  */

   DeleteParallelLock(&lockid);

/*  delete parallel events  */

   DeleteParallelEvent(&eventa);
   DeleteParallelEvent(&eventb);

/*  output  */

   printf("     x[  0]= %2.4f\n", x[  0]);
   printf("     x[ 49]= %2.4f\n", x[ 49]);
   printf("     x[ 99]= %2.4f\n", x[ 99]);
   printf("     x[149]= %2.4f\n", x[149]);
   printf("     x[199]= %2.4f\n", x[199]);
}
      





void SubA( float a[order][order],
           float x[order],
           float b[order],
           int *n,
           float r[order],
           float p[order],
           float workold[numberoftasks],
           float worknew[numberoftasks],
           float workpap[numberoftasks],
           float *alpha,
           float *beta,
           int *tasks,
           int *lockid,
           int *eventa,
           int *eventb,
           int *subset1,
           int *subset23,
           int *subset4,
           int *subset56,
           int *subset78,
           int *subset9,
           int *outstanding,
           int *taskindex,
           float *tolerance,
           int *done)
{
/*
;
;  function to solve a system of linear equations in the form
;  [A]{X}={B} by the Conjugate gradient method, in which [A] is a
;  positive definite matrix of order (NxN), {B} is the right side
;  vector, and {X} is the solution to be determined.
;  (A)Parameters:
;     1.a: left side matrix, dimension[n][n]
;     2.x: return solution, dimension[n]
;     3.b: right side vector, dimension[n]
;     4.n: order
;     5.r: working array for residual, dimension[n]
;     6.p: working array for directional vector, dimension[n]
;     7.workold: working array for the residual norm in
;                previous iteration, dimension[tasks]
;     8.worknew: working array for the residual norm in
;                current iteration, dimension[tasks]
;     9.workpap: working array for {P}[A]{P}, dimension[tasks]
;    10.alpha,beta: working array for the coefficients alpha and beta
;    11.tasks: number of tasks
;    12.lockid: lock identifier
;    13.eventa: identifier of the first event, which must be
;               initialized in the caller's program
;    14.eventb: identifier of the second event
;    15.subset1: length of subset for computing function 1
;    16.subset23: length of subset for computing functions 2 and 3
;    17.subset4: length of subset for computing function 4
;    18.subset56: length of subset for computing functions 5 and 6
;    19.subset78: length of subset for computing functions 7 and 8
;    20.subset9: length of subset for computing function 9
;    21.outstanding: outstanding element that must be initialized
;                         to one in the caller
;    22.taskIndex: task ordinal number
;    23.tolerance: tolerance to terminate computations
;    24.done: working variable for termination of computation
;
*/

/*  private variables  */

   int j4temp$, k4temp$, m4temp$, yes$;

/*  set initial value  */

   *done = 0;
/*
;
;  [critical section] --- function 1
;  distribute subsets
;
*/
label_10:    
   LockonParallelLock(lockid);
   if( *outstanding > *n - 1 )
   {
      UnlockParallelLock(lockid);
      goto label_20;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subset1;
   UnlockParallelLock(lockid);
/*
;
;  [parallelism] --- function 1
;  initialize solution vector, Size=subset1
;
*/
   for ( k4temp$ = j4temp$;
         k4temp$ < __min( j4temp$ + *subset1, *n );
         k4temp$ ++ )
         x[k4temp$] = 0.0;
   goto label_10;
/*
;
;  [task synchronization] --- function 1
;  complete a cycle of parallel event A
;
*/
label_20:    
   PostParallelEvent(eventa);
   IfCompleteParallelEvent(eventa, &yes$);
/*
;
;  [sequential fragment] --- function 1
;  initialize parallel event B and oustanding element
;
*/
   if( yes$ == 0)
   {
      InitializeParallelEvent(eventb, tasks);
      *outstanding = 0;
      CompleteParallelEvent(eventa);
   }
/*
;
;  [critical section] --- functions 2 and 3
;  distribute subsets
;
*/
   workold[*taskindex] = 0.0;
label_30:    
   LockonParallelLock(lockid);
   if( *outstanding > *n - 1 )
   {
      UnlockParallelLock(lockid);
      goto label_40;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subset23;
   UnlockParallelLock(lockid);
/*
;
;  [parallelism] --- functions 2 and 3
;  compute residual and norm, Size=4 * subset23
;
*/
   for (k4temp$ = j4temp$;
        k4temp$ < __min( j4temp$ + *subset23, *n);
        k4temp$ ++ )
   {
      r[k4temp$] = b[k4temp$];
      p[k4temp$] = r[k4temp$];
      workold[*taskindex] = workold[*taskindex] + r[k4temp$] * r[k4temp$];
   }
   goto label_30;
/*
;
;  [task synchronization] --- functions 2 and 3
;  complete a cycle of event B
;
*/
label_40:    
   PostParallelEvent(eventb);
   IfCompleteParallelEvent(eventb, &yes$);
/*
;
;  [sequential fragment] --- functions 2 and 3
;  sum the residual norm, and initialize event A and outstanding element
;
*/
   if( yes$ == 0 )
   {
      for ( j4temp$ = 1; j4temp$ < *tasks; j4temp$ ++ )
            workold[0] = workold [0] + workold[j4temp$];
      InitializeParallelEvent(eventa, tasks);
      *outstanding = 0;
      CompleteParallelEvent(eventb);
   }
/*
;
;  [critical section] --- function 4
;  distribute subsets
;
*/
   workpap[*taskindex] = 0.0;
label_50:    
   LockonParallelLock(lockid);
   if( *outstanding > *n - 1)
   {
      UnlockParallelLock(lockid);
      goto label_60;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subset4;
   UnlockParallelLock(lockid);
/*
;
;  [parallelism] --- function 4
;  compute a partial sum of {P}[A]{P}, size=3 * n * subset4
;
*/
   for ( k4temp$ = j4temp$;
         k4temp$ < __min( j4temp$ + *subset4, *n );
         k4temp$ ++ )
         for (m4temp$ = 0; m4temp$ < *n; m4temp$ ++)
            workpap[*taskindex] = workpap[*taskindex] + p[k4temp$] *
                                                        a[k4temp$][m4temp$] *
                                                        p[m4temp$];
   goto label_50;
/*
;
;  [task synchronization] --- function 4
;  complete a cycle of event A
;
*/
label_60:    
   PostParallelEvent(eventa);
   IfCompleteParallelEvent(eventa, &yes$);
/*
;
;  [sequential fragment] --- function 4
;  sum {P}[A]{P}, initialize event B and outstanding element,
;  compute coefficient alpha
;
*/
   if( yes$ == 0 )
   {
      for ( j4temp$ = 1; j4temp$ < *tasks; j4temp$ ++ )
            workpap[0] = workpap[0] + workpap[j4temp$];
      InitializeParallelEvent( eventb, tasks);
      *outstanding = 0;
      *alpha = - workold[0] / workpap[0];
      CompleteParallelEvent(eventa);
   }
/*
;
;  [critical section] --- functions 5 and 6
;  distribute subsets
;
*/
label_70:    
   LockonParallelLock(lockid);
   if( *outstanding > *n - 1 )
   {
      UnlockParallelLock(lockid);
      goto label_80;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subset56;
   UnlockParallelLock(lockid);
/*
;
;  [parallellism] --- functions 5 and 6
;  compute a partial solution, and residual, size= subset56 * (3 * n + 2)
;
*/
   for ( k4temp$ = j4temp$;
         k4temp$ < __min( j4temp$ + *subset56, *n );
         k4temp$ ++ )
   {
      x[k4temp$] = x[k4temp$] - *alpha * p[k4temp$];
      for ( m4temp$ = 0; m4temp$ < *n; m4temp$ ++ )
            r[k4temp$] = r[k4temp$] + *alpha * 
                                      a[k4temp$][m4temp$] * 
                                      p[m4temp$];
   }
   goto label_70;
/*
;
;  [task synchronization] --- functions 5 and 6
;  complete a cycle of event B
;
*/
label_80:    
   PostParallelEvent(eventb);
   IfCompleteParallelEvent(eventb, &yes$);
/*
;
;  [sequential fragment] --- functions 5 and 6
;  initialize event A and outstanding element
;
*/
   if( yes$ == 0)
   {
      InitializeParallelEvent(eventa, tasks);
      *outstanding = 0;
      CompleteParallelEvent(eventb);
   }
/*
;
;  [critical section] --- functions 7 and 8
;  distribute subset
;
*/
   worknew[*taskindex] = 0.0;
label_90:    
   LockonParallelLock(lockid);
   if( *outstanding > *n - 1 )
   {
      UnlockParallelLock(lockid);
      goto label_100;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subset78;
   UnlockParallelLock(lockid);
/*
;
;  [parallelism] --- functions 7 and 8
;  compute a partial sum of norm, size=2 * subset78
;
*/
   for ( k4temp$ = j4temp$;
         k4temp$ < __min( j4temp$ + *subset78, *n );
         k4temp$ ++ )
         worknew[*taskindex] = worknew[*taskindex] + r[k4temp$] *
                                                     r[k4temp$];
   goto label_90;
/*
;
;  [task synchronization]
;  complete a cycle of event A
;
*/
label_100:   
   PostParallelEvent(eventa);
   IfCompleteParallelEvent(eventa, &yes$);
/*
;
;  [sequential fragment]
;  sum residual,
;  and compute coefficient beta and initialize event B if necessary
;
*/
   if( yes$ == 0 )
   {
      for ( j4temp$ = 1; j4temp$ < *tasks; j4temp$ ++ )
            worknew[0] = worknew[0] + worknew[j4temp$];
      if( worknew[0] < *tolerance)
      {
         *done = 1;
      }
      else
      {
         *beta = worknew[0] / workold[0];
         workold[0] = worknew[0];
      }
      *outstanding = 0;
      InitializeParallelEvent(eventb, tasks);
      CompleteParallelEvent(eventa);
   }
/*
;
;  exit the procedure
;
*/
   if( *done == 1) return;
/*   
;
;  [critical section] --- function 9
;  distribute subsets
;
*/
label_110:   
   LockonParallelLock(lockid);
   if( *outstanding > *n - 1 )
   {
      UnlockParallelLock(lockid);
      goto label_120;
   }
   j4temp$ = *outstanding;
   *outstanding = *outstanding + *subset9;
   UnlockParallelLock(lockid);
/*
;
;  [parallelism] --- function 9
;  compute directional vector, size=2 * subset9
;
*/
   for ( k4temp$ = j4temp$ ;
         k4temp$ < __min( j4temp$ + *subset9, *n );
         k4temp$ ++ )
         p[k4temp$] = r[k4temp$] + *beta * p[k4temp$];
   goto label_110;
/*
;
;  [task synchronization] --- function 9
;  complete a cycle of event B
;
*/
label_120:   
   PostParallelEvent(eventb);
   IfCompleteParallelEvent(eventb, &yes$);
/*
;
;  [sequential fragment] --- function 10
;  initialize event A and outstanding element
;
*/
   if( yes$ == 0 )
   {
      InitializeParallelEvent(eventa, tasks);
      *outstanding = 0;
      CompleteParallelEvent(eventb);
   }
/*
;
;  advance to the next iteration
;
*/
   workpap[*taskindex] = 0.0;
   goto label_50;
}
#define m 2147483647
#define a 16807
float crandom(int *seed)
{
   double rtempa, rtempb;
   float rtemp;
   int itemp;
   rtempa = (double) *seed * a;
   itemp = (int) (rtempa / (double) m);
   rtempb = (double) m * (double) itemp;
   *seed = (int) (rtempa - rtempb);
   rtemp = (double) *seed / (double) m;
   return rtemp;
}
